package net.rumati.logging.muffero;

import java.util.LinkedList;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import net.rumati.logging.muffero.appender.filter.PriorityFilterAppender;
import net.rumati.logging.muffero.util.ReadWriteLockedList;

public class Logger
{
    private static final String FQCN = Logger.class.getName();

    private final ReadWriteLockedList<PriorityFilterAppender> appenders = new ReadWriteLockedList<PriorityFilterAppender>(new LinkedList<PriorityFilterAppender>());
    private final String name;

    private final ReadWriteLock lock = new ReentrantReadWriteLock(true);
    private boolean inherit = true;
    private Logger parent;

    Logger(Logger parent, String name)
    {
        this.parent = parent;
        this.name = name;
    }

    public String getName()
    {
        return name;
    }

    public void setInheritParentAppenders(boolean inherit)
    {
        lock.writeLock().lock();
        try {
            this.inherit = inherit;
        }finally{
            lock.writeLock().unlock();
        }
    }

    public boolean inheritsParentAppenders()
    {
        lock.readLock().lock();
        try {
            return inherit;
        }finally{
            lock.readLock().unlock();
        }
    }

    void addAppender(PriorityFilterAppender appender)
    {
        appenders.add(appender);
    }

    void removeAppender(PriorityFilterAppender appender)
    {
        appenders.remove(appender);
    }

    Logger getParent()
    {
        lock.readLock().lock();
        try {
            return parent;
        }finally{
            lock.readLock().unlock();
        }
    }

    void setParent(Logger parent)
    {
        lock.writeLock().lock();
        try {
            this.parent = parent;
        }finally{
            lock.writeLock().unlock();
        }
    }

    private boolean isPriorityEnabled(Priority p)
    {
        appenders.readLock();
        try {
            for (PriorityFilterAppender appender : appenders){
                if (appender.isLevelEnabled(p)){
                    return true;
                }
            }
        }finally{
            appenders.readUnlock();
        }
        if (inheritsParentAppenders()){
            return getParent().isPriorityEnabled(p);
        }else{
            return false;
        }
    }

    public boolean isTraceEnabled()
    {
        return isPriorityEnabled(Priority.TRACE);
    }

    public void trace(String msg)
    {
        trace(msg, null);
    }

    public void trace(String msg, Throwable t)
    {
        log(Priority.TRACE, FQCN, msg, t);
    }

    public boolean isDebugEnabled()
    {
        return isPriorityEnabled(Priority.DEBUG);
    }

    public void debug(String msg)
    {
        debug(msg, null);
    }

    public void debug(String msg, Throwable t)
    {
        log(Priority.DEBUG, FQCN, msg, t);
    }

    public boolean isInfoEnabled()
    {
        return isPriorityEnabled(Priority.INFO);
    }

    public void info(String msg)
    {
        info(msg, null);
    }

    public void info(String msg, Throwable t)
    {
        log(Priority.INFO, FQCN, msg, t);
    }

    public boolean isWarnEnabled()
    {
        return isPriorityEnabled(Priority.WARNING);
    }

    public void warn(String msg)
    {
        warn(msg, null);
    }

    public void warn(String msg, Throwable t)
    {
        log(Priority.WARNING, FQCN, msg, t);
    }

    public boolean isErrorEnabled()
    {
        return isPriorityEnabled(Priority.ERROR);
    }

    public void error(String msg)
    {
        error(msg, null);
    }

    public void error(String msg, Throwable t)
    {
        log(Priority.ERROR, FQCN, msg, t);
    }

    public void log(Priority priority, String fqcn, String message, Throwable t)
    {
        if (priority.isMoreCriticalThan(Priority.ERROR)){
            throw new IllegalArgumentException("Cannot log at level: " + priority.name());
        }
        
        LogEvent event = new LogEvent(fqcn, name, priority, message, t);

        log(event);
    }

    private void log(LogEvent event)
    {
        appenders.readLock();
        try {
            for (PriorityFilterAppender appender : appenders){
                appender.append(event);
            }
        }finally{
            appenders.readUnlock();
        }

        if (inheritsParentAppenders()){
            getParent().log(event);
        }
    }

    @Override
    public String toString()
    {
        return "Logger: " + name;
    }
}
