package net.rumati.logging.muffero.config.prop;

import java.io.File;
import java.io.FileNotFoundException;
import java.nio.charset.Charset;
import java.nio.charset.IllegalCharsetNameException;
import java.nio.charset.UnsupportedCharsetException;
import java.util.Map;
import net.rumati.logging.muffero.Layout;
import net.rumati.logging.muffero.appender.ConsoleAppender;
import net.rumati.logging.muffero.appender.DatedFileAppender;
import net.rumati.logging.muffero.appender.FileAppender;
import net.rumati.logging.muffero.LoggingException;
import net.rumati.logging.muffero.layout.PatternLayout;

class DefaultPropertyObjectLoader
    extends PropertyObjectLoaderAggregator
{
    public DefaultPropertyObjectLoader()
    {
        addLoader(new SimplePropertyObjectLoader<PatternLayout>(PatternLayout.class) {

            @Override
            public PatternLayout load(Map<String, String> configuration, PropertyConfigContext ctx)
                    throws LoggingException
            {
                String pattern = configuration.get("pattern");
                if (pattern == null){
                    throw new LoggingException("No pattern provided for Layout: " + PatternLayout.class.getName());
                }
                return new PatternLayout(pattern);
            }
        });

        addLoader(new SimplePropertyObjectLoader<ConsoleAppender>(ConsoleAppender.class) {

            @Override
            public ConsoleAppender load(Map<String, String> configuration, PropertyConfigContext ctx)
                    throws LoggingException
            {
                String layoutClassName = configuration.get("layout");

                if (layoutClassName == null){
                    throw new LoggingException("No layout class provided for ConsoleAppender");
                }

                Layout layout = ctx.loadObjectWithConfigPrefix(layoutClassName, Layout.class, "layout.", configuration);

                return new ConsoleAppender(layout);
            }
        });

        addLoader(new SimplePropertyObjectLoader<FileAppender>(FileAppender.class) {

            @Override
            public FileAppender load(Map<String, String> configuration, PropertyConfigContext ctx)
                    throws LoggingException
            {
                String fileName = configuration.get("file");
                if (fileName == null){
                    throw new LoggingException("No file name configured for " + FileAppender.class.getName());
                }

                String layoutClass = configuration.get("layout");
                if (layoutClass == null){
                    throw new LoggingException("No layout configured for " + FileAppender.class.getName());
                }
                Layout layout = ctx.loadObjectWithConfigPrefix(layoutClass, Layout.class, "layout.", configuration);

                String encoding = configuration.get("encoding");
                try {
                    if (encoding != null){
                        Charset cs = null;
                        try {
                            cs = Charset.forName(encoding);
                        }catch (IllegalCharsetNameException e){
                            throw new LoggingException("Invalid encoding for " + FileAppender.class.getName(), e);
                        }catch (UnsupportedCharsetException e){
                            throw new LoggingException("Invalid encoding for " + FileAppender.class.getName(), e);
                        }

                        return new FileAppender(new File(fileName), cs, layout);
                    }else{
                        return new FileAppender(new File(fileName), layout);
                    }
                }catch (FileNotFoundException e){
                    throw new LoggingException("Invalid file for " + FileAppender.class.getName(), e);
                }
            }
        });

        addLoader(new SimplePropertyObjectLoader<DatedFileAppender>(DatedFileAppender.class) {

            @Override
            public DatedFileAppender load(Map<String, String> configuration, PropertyConfigContext ctx)
                    throws LoggingException
            {
                String fileNamePattern = configuration.get("fileNamePattern");
                if (fileNamePattern == null){
                    throw new LoggingException("No file name pattern configured for " + DatedFileAppender.class.getName());
                }

                String layoutClass = configuration.get("layout");
                if (layoutClass == null){
                    throw new LoggingException("No layout configured for " + DatedFileAppender.class.getName());
                }
                Layout layout = ctx.loadObjectWithConfigPrefix(layoutClass, Layout.class, "layout.", configuration);

                Charset cs = null;
                String encoding = configuration.get("encoding");
                if (encoding != null){
                    cs = null;
                    try {
                        cs = Charset.forName(encoding);
                    }catch (IllegalCharsetNameException e){
                        throw new LoggingException("Invalid encoding for " + DatedFileAppender.class.getName(), e);
                    }catch (UnsupportedCharsetException e){
                        throw new LoggingException("Invalid encoding for " + DatedFileAppender.class.getName(), e);
                    }
                }

                try {
                    if (cs == null){
                        return new DatedFileAppender(fileNamePattern, layout);
                    }else{
                        return new DatedFileAppender(fileNamePattern, cs, layout);
                    }
                }catch (IllegalArgumentException e){
                    throw new LoggingException("Invalid fileNamePattern for " + DatedFileAppender.class.getName(), e);
                }
            }
        });
    }
}
