package net.rumati.logging.muffero.config.prop;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;
import net.rumati.logging.muffero.Appender;
import net.rumati.logging.muffero.Configurator;
import net.rumati.logging.muffero.Priority;
import net.rumati.logging.muffero.Logger;
import net.rumati.logging.muffero.LoggerFactory;
import net.rumati.logging.muffero.LoggingException;
import net.rumati.logging.muffero.util.ClassUtil;

public class PropertyConfigurator
    extends Configurator
{
    private final List<PropertyObjectLoader> loaders = new LinkedList<PropertyObjectLoader>();
    private final Map<String, String> config = new TreeMap<String, String>();

    public PropertyConfigurator(Properties properties)
    {
        this(propertiesToMap(properties));
    }

    private static Map<String, String> propertiesToMap(Properties properties)
    {
        Map<String, String> config = new TreeMap<String, String>();
        for (Map.Entry<Object, Object> e : properties.entrySet()){
            Object key = e.getKey();
            Object value = e.getValue();

            if (key != null && value != null){
                config.put(key.toString(), value.toString());
            }
        }
        return config;
    }

    public PropertyConfigurator(Map<String, String> config)
    {
        loaders.add(new DefaultPropertyObjectLoader());
        this.config.putAll(config);
        for (Map.Entry<String, String> e : this.config.entrySet()){
            for (Map.Entry<String, String> e2 : this.config.entrySet()){
                if (!e2.getKey().equals(e.getKey())){
                    e2.setValue(e2.getValue().replace("${" + e.getKey() + "}", e.getValue()));
                }
            }
        }
    }

    public void addObjectLoader(PropertyObjectLoader loader)
    {
        loaders.add(loader);
    }

    public void configure()
            throws LoggingException
    {
        String loaderList = config.get("muffero.loaders");
        while (loaderList != null){
            String loaderClassName;
            int idx = loaderList.indexOf(",");
            if (idx < 0){
                loaderClassName = loaderList;
                loaderList = null;
            }else{
                loaderClassName = loaderList.substring(0, idx).trim();
                loaderList = loaderList.substring(idx + 1).trim();
            }

            try{
                Class c = Class.forName(loaderClassName);
                if (!ClassUtil.classImplements(c, PropertyObjectLoader.class)){
                    throw new LoggingException("Class \"" + c.getName() + "\" does not implement class \""
                            + PropertyObjectLoader.class.getName() + "\"");
                }
                try{
                    PropertyObjectLoader l = (PropertyObjectLoader)c.newInstance();
                    loaders.add(l);
                }catch (IllegalAccessException e){
                    throw new LoggingException("Error instantiating object loader " + loaderClassName, e);
                }catch (InstantiationException e){
                    throw new LoggingException("Error instantiating object loader " + loaderClassName, e);
                }catch (ExceptionInInitializerError e){
                    throw new LoggingException("Error instantiating object loader " + loaderClassName, e);
                }
            }catch (ClassNotFoundException e){
                throw new LoggingException("Unable to load object loader class: " + loaderClassName);
            }
        }

        PropertyConfigContext ctx = new PropertyConfigContext(loaders);

        Map<String, Appender> appenders = new TreeMap<String, Appender>();
        for (Map.Entry<String, String> e : config.entrySet()){
            String key = e.getKey();
            String prefix = "muffero.appender.";
            if (key.startsWith(prefix) && key.lastIndexOf(".") < prefix.length()){
                String appenderName = key.substring(prefix.length());
                Appender appender = ctx.loadObjectWithConfigPrefix(e.getValue(), Appender.class, key + ".", config);
                appenders.put(appenderName, appender);
            }
        }

        for (Map.Entry<String, String> configEntry : config.entrySet()){
            String prefix = "muffero.logger.";
            String key = configEntry.getKey();
            if (key.startsWith(prefix) && key.length() > prefix.length()){
                String loggerName = key.substring(prefix.length());
                loadLoggerConfig(LoggerFactory.getLogger(loggerName), configEntry.getValue(), appenders);
            }
        }

        String rootLoggerConfig = config.get("muffero.rootLogger");
        if (rootLoggerConfig != null){
            loadLoggerConfig(LoggerFactory.getRootLogger(), rootLoggerConfig, appenders);
        }
    }


    private void loadLoggerConfig(Logger logger, String configuration, Map<String, Appender> appenders) throws LoggingException
    {
        while (configuration != null){
            String item;
            int idx = configuration.indexOf(",");
            if (idx < 0){
                item = configuration;
                configuration = null;
            }else{
                item = configuration.substring(0, idx).trim();
                configuration = configuration.substring(idx + 1).trim();
            }

            if (item.equals("NOINHERIT")){
                logger.setInheritParentAppenders(false);
            }else{
                idx = item.indexOf(":");
                if (idx < 0){
                    throw new LoggingException("Invalid configuration item \"" + item + "\" for logger \"" + logger.getName() + "\"");
                }
                Priority priority = null;
                try{
                    priority = Priority.valueOf(item.substring(0, idx));
                }catch (IllegalArgumentException e){
                    throw new LoggingException("Invalid configuration item \"" + item + "\" for logger \"" + logger.getName() + "\"");
                }
                String appenderName = item.substring(idx + 1);
                Appender appender = appenders.get(appenderName);
                if (appender == null){
                    throw new LoggingException("Unknown appender \"" + appenderName + "\" for logger \"" + logger.getName() + "\"");
                }
                addAppender(logger, appender, priority);
            }
        }
    }
}
