package net.rumati.logging.muffero.layout;

import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Map.Entry;
import net.rumati.logging.muffero.LogEvent;
import net.rumati.logging.muffero.MDC;
import net.rumati.logging.muffero.util.FastDateFormat;

class PatternParser
{
    private static interface Directive
    {
        public String[] getNames();

        public DirectiveFormatter getFormatter(String parameter);
    }

    private static class FormatModifiers
    {
        private int minWidth = 0;
        private int maxWidth = 0;
        private boolean leftAligned = false;
        private boolean truncatedAtEnd = false;
    }

    public static interface Formatter
    {
        public String getFormattedText(LogEvent event);
    }

    private static class StaticTextFormatter
            implements Formatter
    {
        private final String text;

        public StaticTextFormatter(String test)
        {
            this.text = test;
        }

        public String getFormattedText(LogEvent event)
        {
            return text;
        }
    }

    private static interface DirectiveFormatter
            extends Formatter
    {
        public void setModifiers(FormatModifiers modifiers);
    }

    private static abstract class ModifierIgnoringDirectiveFormatter
            implements DirectiveFormatter
    {
        public void setModifiers(FormatModifiers modifiers)
        {
        }
    }

    private static abstract class ModifiableDireactiveFormatter
            implements DirectiveFormatter
    {
        private FormatModifiers modifiers = null;

        public final void setModifiers(FormatModifiers modifiers)
        {
            this.modifiers = modifiers;
        }

        public final String getFormattedText(LogEvent event)
        {
            String text = getUnmodifiedText(event);
            if (modifiers.maxWidth > 0 && text.length() > modifiers.maxWidth){
                if (modifiers.truncatedAtEnd){
                    return text.substring(0, modifiers.maxWidth);
                }else{
                    return text.substring(text.length() - modifiers.maxWidth);
                }
            }else{
                int add = modifiers.minWidth - text.length();
                if (add > 0){
                    StringBuilder sb = new StringBuilder(add + text.length());
                    if (modifiers.leftAligned){
                        sb.append(text);
                        for (int i = 0; i < add; i++){
                            sb.append(' ');
                        }
                    }else{
                        for (int i = 0; i < add; i++){
                            sb.append(' ');
                        }
                        sb.append(text);
                    }
                    return sb.toString();
                }
                return text;
            }
        }

        public abstract String getUnmodifiedText(LogEvent event);
    }

    private static class DateDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        private final FastDateFormat dateFormat;

        public DateDirectiveFormatter(String datePattern)
        {
            this.dateFormat = FastDateFormat.getInstance(datePattern);
        }

        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return dateFormat.format(event.getDate());
        }
    }

    private static class LoggerNameDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return event.getLoggerName();
        }
    }

    private static class ThreadIdDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return "" + event.getThread().getId();
        }
    }

    private static class ThreadNameDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return event.getThread().getName();
        }
    }

    private static class MessageDirectiveFormatter
            extends ModifierIgnoringDirectiveFormatter
    {
        public String getFormattedText(LogEvent event)
        {
            return event.getMessage();
        }
    }

    private static class RelativeTimeDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        private final long startTime = new Date().getTime();

        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return "" + (event.getTime() - startTime);
        }
    }

    private static class NewLineDirectiveFormatter
            extends ModifierIgnoringDirectiveFormatter
    {
        private static final String newLine = System.getProperty("line.separator");

        public String getFormattedText(LogEvent event)
        {
            return newLine;
        }
    }

    private static class PriorityDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return event.getPriority().name();
        }
    }

    private static class CallerLocationDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return event.getStackTrace()[0].toString();
        }
    }

    private static class CallerClassDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return event.getStackTrace()[0].getClassName();
        }
    }

    private static class CallerSourceFileDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return event.getStackTrace()[0].getFileName();
        }
    }

    private static class CallerSourceLineDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return "" + event.getStackTrace()[0].getLineNumber();
        }
    }

    private static class CallerMethodNameDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            return event.getStackTrace()[0].getMethodName();
        }
    }

    private static class IndexedMDCDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        private final String key;

        public IndexedMDCDirectiveFormatter(String key)
        {
            this.key = key;
        }

        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            String value = MDC.get(key);
            if (value == null){
                return "";
            }
            return value;
        }
    }

    private static class CompleteMDCDirectiveFormatter
            extends ModifiableDireactiveFormatter
    {
        @Override
        public String getUnmodifiedText(LogEvent event)
        {
            StringBuilder sb = new StringBuilder(100);
            boolean first = true;
            for (Entry<String, String> e : MDC.getCopyOfContextMap().entrySet()){
                if (first){
                    first = false;
                }else{
                    sb.append(", ");
                }
                sb.append(e.getKey());
                sb.append("=");
                sb.append(e.getValue());
            }
            return sb.toString();
        }
    }

    private static class StackTraceDirectiveFormatter
            extends ModifierIgnoringDirectiveFormatter
    {
        private static final String newLine = System.getProperty("line.separator");
        private final int depth;

        public StackTraceDirectiveFormatter(int depth)
        {
            this.depth = depth;
        }

        public String getFormattedText(LogEvent event)
        {
            StringBuilder sb = new StringBuilder(100);
            if (depth < 1){
                for (StackTraceElement e : event.getStackTrace()){
                    sb.append("    at ");
                    sb.append(e.toString());
                    sb.append(newLine);
                }
            }else{
                StackTraceElement[] st = event.getStackTrace();
                int limit = depth;
                if (limit > st.length){
                    limit = st.length;
                }
                for (int i = 0; i < limit; i++){
                    sb.append("    at ");
                    sb.append(st[i].toString());
                    sb.append(newLine);
                }
                if (limit < st.length){
                    sb.append("    ... ");
                    sb.append(st.length - limit);
                    sb.append(" more");
                }
            }
            return sb.toString();
        }
    }

    private static class ThrowableDirectiveFormatter
            extends ModifierIgnoringDirectiveFormatter
    {
        private static final String newLine = System.getProperty("line.separator");
        private final int depth;

        public ThrowableDirectiveFormatter(int depth)
        {
            this.depth = depth;
        }

        public String getFormattedText(LogEvent event)
        {
            StringBuilder sb = new StringBuilder(100);

            boolean first = true;
            for (Throwable t = event.getThrowable(); t != null; t = t.getCause()){
                if (first){
                    first = false;
                }else{
                    sb.append("Caused by: ");
                }
                sb.append(t.toString());
                sb.append(newLine);
                if (depth == 0){
                    for (StackTraceElement e : t.getStackTrace()){
                        sb.append("    at ");
                        sb.append(e.toString());
                        sb.append(newLine);
                    }
                }else if (depth > 0){
                    StackTraceElement[] st = t.getStackTrace();
                    int limit = depth;
                    if (limit > st.length){
                        limit = st.length;
                    }
                    for (int i = 0; i < limit; i++){
                        sb.append("    at ");
                        sb.append(st[i].toString());
                        sb.append(newLine);
                    }
                    if (limit < st.length){
                        sb.append("    ... ");
                        sb.append(st.length - limit);
                        sb.append(" more");
                    }
                }
            }
            return sb.toString();
        }
    }
    private static final Directive[] directives = new Directive[]{
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "date", "d" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                if ("".equals(parameter) || "ISO8601".equals(parameter)){
                    return new DateDirectiveFormatter("yyyy-MM-dd HH:mm:ss.SSS");
                }
                return new DateDirectiveFormatter(parameter);
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "logger" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new LoggerNameDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "thread" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                if ("id".equals(parameter)){
                    return new ThreadIdDirectiveFormatter();
                }else if ("".equals(parameter) || "name".equals(parameter)){
                    return new ThreadNameDirectiveFormatter();
                }else{
                    throw new IllegalArgumentException("Unknown parameter for thread formatting directive: " + parameter);
                }
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "message", "msg" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new MessageDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "relative", "r" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new RelativeTimeDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "n" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new NewLineDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "level", "priority" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new PriorityDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{"caller"};
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new CallerLocationDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "class", "C" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new CallerClassDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "method", "M" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new CallerMethodNameDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "line", "L" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new CallerSourceLineDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "file", "F" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                return new CallerSourceFileDirectiveFormatter();
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "mdc", "X" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                if (parameter.length() == 0){
                    return new CompleteMDCDirectiveFormatter();
                }
                return new IndexedMDCDirectiveFormatter(parameter);
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "stack" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                if (parameter.length() == 0){
                    return new StackTraceDirectiveFormatter(0);
                }
                try{
                    return new StackTraceDirectiveFormatter(Integer.valueOf(parameter));
                }catch (NumberFormatException e){
                    throw new IllegalArgumentException("Invalid depth parameter for %stack formatting directive: " + parameter, e);
                }
            }
        },
        new Directive()
        {
            public String[] getNames()
            {
                return new String[]{ "throwable", "exception", "ex" };
            }

            public DirectiveFormatter getFormatter(String parameter)
            {
                if (parameter.length() == 0){
                    return new ThrowableDirectiveFormatter(0);
                }else{
                    try{
                        return new ThrowableDirectiveFormatter(Integer.valueOf(parameter));
                    }catch (NumberFormatException e){
                        throw new IllegalArgumentException("Invalid depth parameter for %exception formatting directive: "
                                + parameter, e);
                    }
                }
            }
        }
    };
    private final String pattern;
    private int cur;

    public PatternParser(String pattern)
    {
        this.pattern = pattern;
    }

    public List<Formatter> parse()
    {
        List<Formatter> formatters = new LinkedList<Formatter>();
        int length = pattern.length();
        StringBuilder text = new StringBuilder();
        for (cur = 0; cur < length; cur++){
            char c;
            if ((c = pattern.charAt(cur)) != '%'){
                text.append(c);
            }else{
                /*
                 * Check that this is not the last character of the pattern. If so, just break here,
                 * as this cant possible be a conversion specifier. We must just treat it as a normal
                 * % sign.
                 */
                if (cur == length - 1){
                    text.append('%');
                    break;
                }

                /*
                 * Check for %%, which we treat as just a single %
                 */
                if (pattern.charAt(cur + 1) == '%'){
                    text.append('%');
                    cur++;
                    continue;
                }

                /*
                 * Add appender with existing plain text
                 */
                if (text.length() > 0){
                    formatters.add(new StaticTextFormatter(text.toString()));
                }

                /*
                 * Start reading directive from after the percent sign
                 */
                cur++;

                formatters.add(readDirective());

                text = new StringBuilder();
            }
        }
        if (text.length() > 0){
            formatters.add(new StaticTextFormatter(text.toString()));
        }
        return formatters;
    }

    private FormatModifiers readFormatModifiers()
    {
        FormatModifiers modifiers = new FormatModifiers();
        if (pattern.charAt(cur) == '-'){
            modifiers.leftAligned = true;
            cur++;
        }
        int start = cur;
        while (cur < pattern.length() && Character.isDigit(pattern.charAt(cur))){
            cur++;
        }
        if (cur > start){
            modifiers.minWidth = Integer.valueOf(pattern.substring(start, cur));
        }
        if (cur < pattern.length() - 1 && pattern.charAt(cur) == '.'){
            cur++;

            if (pattern.charAt(cur) == '-'){
                modifiers.truncatedAtEnd = true;
                cur++;
            }

            start = cur;
            while (cur < pattern.length() && Character.isDigit(pattern.charAt(cur))){
                cur++;
            }
            if (cur > start){
                modifiers.maxWidth = Integer.valueOf(pattern.substring(start, cur));
            }
        }

        return modifiers;
    }

    private String readDirectiveParameter()
    {
        int length = pattern.length();
        if (cur < length && pattern.charAt(cur) == '{'){
            cur++;
            int start = cur;
            int brackets = 1;
            for (; cur < length; cur++){
                if (pattern.charAt(cur) == '{'){
                    brackets++;
                }else if (pattern.charAt(cur) == '}'){
                    brackets--;
                    if (brackets == 0){
                        return pattern.substring(start, cur);
                    }
                }
            }
            throw new IllegalArgumentException("Unmatched brace in pattern");
        }else{
            cur--;
        }
        return "";
    }

    private DirectiveFormatter readDirective()
    {
        FormatModifiers formatModifiers = readFormatModifiers();

        String fromStartOfDirectiveName = pattern.substring(cur);

        for (Directive directive : directives){
            for (String directiveName : directive.getNames()){
                if (fromStartOfDirectiveName.startsWith(directiveName)){
                    cur += directiveName.length();
                    String parameter = readDirectiveParameter();
                    DirectiveFormatter formatter = directive.getFormatter(parameter);
                    formatter.setModifiers(formatModifiers);
                    return formatter;
                }
            }
        }

        int idx = fromStartOfDirectiveName.indexOf("%");
        if (idx >= 0){
            fromStartOfDirectiveName = fromStartOfDirectiveName.substring(0, idx);
        }
        idx = fromStartOfDirectiveName.indexOf(" ");
        if (idx >= 0){
            fromStartOfDirectiveName = fromStartOfDirectiveName.substring(0, idx);
        }
        idx = fromStartOfDirectiveName.indexOf("{");
        if (idx >= 0){
            fromStartOfDirectiveName = fromStartOfDirectiveName.substring(0, idx);
        }

        throw new IllegalArgumentException("Unknown formatting directive in pattern, at: %" + fromStartOfDirectiveName);
    }
}
